/**
 * Core Service
 * Provides access to otherwise global Data which is prepopulated
 * by the administration Backend
 */
angular.module('app').service('core', ['$http', '$controller', '$q', 'message', function($http, $controller, $q, message) {
	
	if (window.dropperCore) return window.dropperCore;
	
	var core = {
		settings: {},
		fn: {
			_listeners: {},

			on: function(evt, cb) {
				if (!core.fn._listeners) {
					core.fn._listeners = {};	
				}
				
				if (!core.fn._listeners[evt]) {
					core.fn._listeners[evt] = [];
				}
				
				// dont allow double registration of the same listener
				if (core.fn._listeners[evt].indexOf(cb) >= 0) return;

				core.fn._listeners[evt].push(cb);
			},
			
			trigger: function(evt, params) {
				if (!core.fn._listeners || !core.fn._listeners[evt]) return [];

				var resList = [];

				for(var i=0; i<core.fn._listeners[evt].length; i++) {
					var cb = core.fn._listeners[evt][i];
					if (typeof(cb) === 'function') {
						resList.push(cb.apply(this, params));
					}
				}
				return resList;
			},
			
			onReload: function(cb) {
				core.fn.on('reload', cb);
			},
			
			onReloading: function(cb) {
				core.fn.on('reloading', cb);
			},

			loadSettings: function() {
				var defer = $q.defer();
				$.ajax({
					type: 'POST',
					dataType: 'JSON',
					url: window.kk_shop_url,
					data: {
						"kk_dropper_settings": true,
						"kk_dropper_admin_pass": window.kk_admin_secret
					},
					success: function(res) {
						if (core.settings) delete core.settings;
						core.settings = res;
						core.latestRevision = core.settings.meta;
						defer.resolve(core.settings);
					},
					error: function(err) {
						defer.reject(err);
						message.errorHTML('Fehler beim Laden der Dropper Einstellungen!', '&bull;&nbsp;<a href="https://kreativkonzentrat.de/Wiki?page=dropper/faq" target="_blank">Troubleshooting & FAQs</a><br>&bull;&nbsp;<a href="https://kreativkonzentrat.de/Dropper?ticket" target="_blank">Support-Ticket erstellen</a>');
						message.error('Fehler beim Laden der Dropper Einstellungen!', err);
					}
				});
				return defer.promise;
			},

			reload: function() {
				var defer = $q.defer();
				
				$.ajax({
					type: 'POST',
					dataType: 'JSON',
					url: window.kk_shop_url,
					data: {
						"kk_dropper_core_data": true,
						"kk_dropper_admin_pass": window.kk_admin_secret
					},
					success: function(res) {
						// remove old properties
						for(var prop in core) {
							if (prop !== 'fn' && prop !== 'settings' && prop !== 'latestRevision') delete core[prop];
						}
			
						// set the new properties
						$.extend(true, core, res);
	
						// set admin user
						core.adminUser = window.kk_admin_user;
	
						// trigger reloading event. the callbacks will return promises so that we will wait for them
						// to finish, before we send out the final core resolve
						var beforePromises = core.fn.trigger('reloading', [core]);
	
						if (beforePromises && beforePromises.length) {
							$q.all(beforePromises).then(function() {
								// set resolved
								defer.resolve(core);
								// trigger the reload event
								core.fn.trigger('reload', [core]);
							});
						} else {
							// set resolved
							defer.resolve(core);
							// trigger the reload event
							core.fn.trigger('reload', [core]);
						}
					},
					error: function(err) {
						defer.reject(err);
						message.errorHTML('Fehler bei der Dropper Initialisierung!', '&bull;&nbsp;<a href="https://kreativkonzentrat.de/Wiki?page=dropper/faq" target="_blank">Troubleshooting & FAQs</a><br>&bull;&nbsp;<a href="https://kreativkonzentrat.de/Dropper?ticket" target="_blank">Support-Ticket erstellen</a>');
						message.error('Fehler bei der Dropper Initialisierung!', err);
					}
				});

				return defer.promise;
			}
		}
	};
	
	window.dropperCore = core;
	return window.dropperCore;
	
}]);

/**
 * Message Service
 * provides basic functionality for notifiying the user
 * There are 5 types of messages: error, warning, info, success
 * All of them take a message string, an object (output as json) and optionally a delay time
 */
angular.module('app').service('message', ['$rootScope', function($rootScope) {

	// check if we were have been initialized already...
	if (window.dropperMessageService) return window.dropperMessageService;

	// build service instance	
	var msgService = {
	
		// list of shown messages
		messages: [],

		// internal message function
		_message: function(type, title, body, time, allowHTML) {
	
			var newMsg = {
				type: type,
				title: title,
				body: typeof body === 'string' ? body : JSON.stringify(body),
				time: time,
				created: new Date(),
				allowHTML: allowHTML ? true : false
			};
	
			this.messages.push(newMsg);
	
			if (time !== 0) {
				var self = this;
				setTimeout(function() {
					$rootScope.$apply(function() {
						self.removeMessage(newMsg);
					});
				}, time);
			}
	
			if (!$rootScope.$$phase) $rootScope.$apply();
		},
	
		// send error message
		error: function(title, body, time) {
			this._message('error', title, body, (typeof(time) == 'number' ? time : 0), false);
			console.log('Error: ', title, body);
		},
	
		// send warning message
		warning: function(title, body, time) {
			this._message('warning', title, body, (typeof(time) == 'number' ? time : 20000), false);
			console.log('Warning: ', title, body);
		},
	
		// send info message
		info: function(title, body, time) {
			this._message('info', title, body, (typeof(time) == 'number' ? time : 6000), false);
		},
		
		// send success message
		success: function(title, body, time) {
			this._message('success', title, body, (typeof(time) == 'number' ? time : 6000), false);
		},
	
		// send error message
		errorHTML: function(title, body, time) {
			this._message('error', title, body, (typeof(time) == 'number' ? time : 0), true);
			console.log('Error: ', title, body);
		},
	
		// send warning message
		warningHTML: function(title, body, time) {
			this._message('warning', title, body, (typeof(time) == 'number' ? time : 20000), true);
			console.log('Warning: ', title, body);
		},
	
		// send info message
		infoHTML: function(title, body, time) {
			this._message('info', title, body, (typeof(time) == 'number' ? time : 6000), true);
		},
		
		// send success message
		successHTML: function(title, body, time) {
			this._message('success', title, body, (typeof(time) == 'number' ? time : 6000), true);
		},
		
	
		// removes a given message from the queue
		removeMessage: function(msg) {
			var id = this.messages.indexOf(msg);
			if (id >= 0) this.messages.splice(id, 1);
		},
	
		// init service
		init: function() {

			// add backend errors on service initialization
			if (window.kk_backend_errors && window.kk_backend_errors.length) {
		
				for (var i=0; i<window.kk_backend_errors.length; i++) {
					var err = window.kk_backend_errors[i];
					// add error
					this.errorHTML(err.title, err.desc || '', 0);
				}
		
				// empty them, because they are added now
				window.kk_backend_errors = [];
		
			}
		}
	};
	
	// initialize
	msgService.init();
	
	// bind to global
	window.dropperMessageService = msgService;
	
	return msgService;

}]);

/**
 * Environement Service
 * Provides functionality about Languages.
 * You can set the current Language here which is later used
 * All inside the UI Bindings
 */
angular.module('app').service('environment', ['core', function(core) {

	if (window.dropperEnvironment) return window.dropperEnvironment;
	
	var env = {
		setLanguage : function(lang) {
			env.language = $.extend(true, {}, lang);
		},
		
		getLanguages: function() {
			return core.languages;
		},
		
		getActiveLang: function() {
			return env.language;
		},
		
		init: function(core) {
			if (core.languages instanceof Array && core.languages.length > 0) {
		
				for(var langIdx in core.languages) {
					// get active and the default language
					if (core.languages[langIdx].cStandard == 'Y') {
						env.language = $.extend(true, {}, core.languages[langIdx]);
						env.languageDefault = $.extend(true, {}, core.languages[langIdx]);
						break;
					}
				}
		
				// no active language could be determined, use the first language available
				if (!env.language) {
					env.language = $.extend(true, {}, core.languages[0]);
				}
		
				// no default language could be determined, use the first language available
				if (!env.languageDefault) {
					env.languageDefault = $.extend(true, {}, core.languages[0]);
				}
			}
			
		}
	};

	core.fn.onReload(env.init);
	env.init(core);

	window.dropperEnvironment = env;
	return window.dropperEnvironment;
}]);

/**
 * Packagemanager Service
 * Provides access to the package manager
 */
angular.module('app').service('packageManager', ['core', function(core) {
	
	var filterByType = function(list, type) {
		if (!type) return list;
		var filtered = {};
		for (var id in list) {
			var pkg = list[id];
			if (pkg && pkg.type === type) filtered[id] = pkg;
		}
		return filtered;
	};
	
	this.hasInstallablePackages = function() {
		return (Object.keys(core.packages.installable).length > 0);
	};

	this.hasUpgradeablePackages = function() {
		return (Object.keys(core.packages.upgradeable).length > 0);
	};
	
	this.hasForcedUpgrades = function() {
		return (Object.keys(this.getForcedUpgrades()).length > 0);
	};

	this.isInstallable = function(pkgId) {
		return (core.packages && core.packages.installable && typeof(core.packages.installable[pkgId]) !== 'undefined');
	};

	this.isUpgradeable = function(pkgId) {
		return (core.packages && core.packages.upgradeable && typeof(core.packages.upgradeable[pkgId]) !== 'undefined');
	};
	
	this.getInstallable = function(type) {
		return filterByType(core.packages.installable, type);
	};

	this.getUpgradeable = function(type) {
		return filterByType(core.packages.upgradeable, type);
	};

	this.getForcedUpgrades = function() {
		var forced = {};
		if (core && core.packages && core.packages.upgradeable) {
			for(var id in core.packages.upgradeable) {
				var upgrade = core.packages.upgradeable[id];
				if (upgrade.candidate && upgrade.candidate.force === true) {
					forced[id] = upgrade;
				}
			}
		}
		return forced;
	};
	
	this.getHoldback = function() {
		return core.packages.holdback;
	};
	
	this.getInstalledVersion = function(pkgId) {
		if (typeof(core.packages.installed[pkgId]) !== 'undefined') {
			return core.packages.installed[pkgId]['version'];
		}
		return false;
	};

}]);

/**
 * Modal Service
 * Allows to create a new Modal
 */
angular.module('app').service('modal', ['$compile', '$rootScope', '$controller', 'core', function($compile, $rootScope, $controller, core) {
	var self = this;

	// DIRTY HACK: Select 2 won't work on modals, this will fix it hopefully
	$.fn.modal.Constructor.prototype.enforceFocus = function() {};

	this.modals = [];

	/**
	 * Closes all open modals
	 */
	this.closeAll = function() {
		for(var i=0; i<this.modals.length; i++) {
			var modal = this.modals[i];
			if (modal && angular.isFunction(modal.close)) {
				modal.close();
			}
		}
	};

	/**
	 * Shows a modal
	 */
	this.show = function(options, callback, injects) {

		// close all other open modals
		this.closeAll();

		// create isolated scope for this modal
		var modalScope = $rootScope.$new(true);
		modalScope.isStatic = false;
		modalScope.delayShow = false;

		if (self.modals.indexOf(modalScope) < 0) self.modals.push(modalScope);

		// let's run this in a timeout, so that we can prevent $digest problems
		setTimeout(function() {
			// $apply is now safe
			modalScope.$apply(function() {
				// indirect call via name
				if (angular.isString(options)) {
					var modalName = options;

					options = {
						template: core.adminURL + 'js/views/modals/' + modalName.toLowerCase() + '.html.js?' + core.cacheID,
						controller: modalName + "Controller"
					};
				}

				// assign the options to our new scope
				modalScope.options = options;

				// compile the template
				var template = '<div class="modal hide fade" tabindex="-1" data-show="false" role="dialog"><ng-include src="options.template" onload="_show()"></ng-include></div>';
				var modalDom = $compile(template)(modalScope);

				// attach the close function to the scope.
				modalScope.close = function(result) {
					if (angular.isFunction(callback)) {
						callback(result);
					}
					$(modalDom).modal('hide');
					var modalIdx = self.modals.indexOf(modalScope);
					// remove the modal from our list
					if (self.modals.indexOf(modalIdx) >= 0) self.modals.splice(modalIdx, 1);
				};

				// show the modal visually
				modalScope.show = function() {
					$(modalDom).modal({
						show: true, 
						backdrop: (modalScope.noBackdrop ? false : modalScope.isStatic ? 'static' : true),
						keyboard: !modalScope.isStatic
					});

					$rootScope.reactivateModals();
				};

				modalScope._show = function() {
					$(modalDom).appendTo('#main-view');

					// create the controller, if any given
					if (options.controller) {
						var ctrlInjects = {"$modalScope": modalScope, "$element": modalDom};
						if (angular.isObject(injects)) {
							ctrlInjects = angular.extend({}, ctrlInjects, injects);
						}
						var controller = $controller(options.controller, ctrlInjects);
					}

					if (!modalScope.delayShow) {
						// show it
						modalScope.show();
					}

					// cleanup when the modal closes
					$(modalDom).on('hidden', function() {
						modalScope.$destroy();
						$(modalDom).remove();
					});
				};
			});
		});
	};
}]);

/**
 * Window Service
 * Allows the creation of Stacked editing Windows
 */
angular.module('app').service('editWindow', ['$compile', '$rootScope', '$controller', '$timeout', 'core', function($compile, $rootScope, $controller, $timeout, core) {
	if (!window.windowHandler) {

		window.windowHandler = {
			// list with all open windows (scopes)
			windows: [],

			// closes all windows
			closeAll: function(cb) {
				var self = this;
				for(var i=0; i<self.windows.length; i++) {
					var wnd = self.windows[i];
					wnd.close(function() {
						if (self.windows.length == 0 && typeof(cb) == 'function') {
							cb();
						}
					}, true);
				}	
			},

			// shows a window
			show: function(nest, options, injects) {
				var self = this;

				// delay the showing of a window so that other windows can close first
				$timeout(function() {
					
					var nested = nest ? true : false;

					var windowScope = $rootScope.$new(true);
					windowScope.nested = nested;
					
					if (angular.isString(options)) {
						options = {
							template: core.adminURL + 'js/views/windows/' + options.toLowerCase() + '.html.js?' + core.cacheID,
							controller: options + "Controller"
						};
					}
	
					windowScope.options = options;

					// called when the templat inside the window is ready and loaded
					windowScope._loadListeners = [];

					windowScope._load = function(cb) {
						if (windowScope._loaded) {
							cb();
						} else {
							windowScope._loadListeners.push(cb);
						}
					};
					
					// when content is loaded call the callbacks
					windowScope.$on('$includeContentLoaded', function () {
						if (windowScope._loaded) return;
						windowScope._loaded = true;

						$timeout(function() {
							for(var i=0; i<windowScope._loadListeners.length; i++) {
								windowScope._loadListeners[i]();
							}
							windowScope._loadListeners = [];
						});
					});

	
					// build the window template
					var template = '<div class="hero-unit edit-window" style="opacity: 0" ng-class="{\'actives\':isActive}" ng-click="activate($event)">\
										<ng-include src="options.template"></ng-include>\
									</div>';
	
					windowScope.dom = $compile(template)(windowScope);

					// make us known to the collection
					if (windowScope.nested) {
						self.windows.push(windowScope);
					} else {
						self.windows.unshift(windowScope);
					}

					// create the controller, if any given
					if (options.controller) {
						var ctrlInjects = {
							"$windowScope": windowScope, 
							"$element": windowScope.dom, 
							"useConditional": true,
							"nameBinding": null
						};

						if (angular.isObject(injects)) {
							ctrlInjects = angular.extend({}, ctrlInjects, injects);
						}
						$controller(options.controller, ctrlInjects);
					}

					// closes the window
					windowScope.close = function(cb, fastClose) {
						// animate out
						windowScope.dom.animate(
							(fastClose ? { 'opacity': '' } : { 'margin-top': '', 'opacity': '' }), 
							(!windowScope.nested || fastClose ? 200 : 400), function() {
							// remove from the dom
							windowScope.dom.remove();
							// destroy our scope
							windowScope.$destroy();
	
							// remove ourself from the windows list
							var idx = self.windows.indexOf(windowScope);
	
							// remove from the window list
							if (idx >= 0) self.windows.splice(idx, 1);
		
							// get the parent, if exists
							var parent = idx > 0 ? self.windows[idx-1] : null;
		
							// activate parent?
							if (parent) {
								parent.activate();
								parent.$apply();
							}
							if (typeof(cb) === 'function') cb(idx);
						});
					};
	
					// sets this window active
					windowScope.activate = function(e) {
						if (windowScope.isActive) return;
						// close all windows higher than ourself
						var wndIdx = self.windows.indexOf(windowScope);
						if (wndIdx >= 0 && self.windows.length > wndIdx) {
							for(var i=wndIdx+1; i<self.windows.length; i++) {
								self.windows[i].close();
							}
						}
						// make all windows inactive
						for(var j=0; j<self.windows.length; j++) {
							self.windows[j].isActive = false;
						}
						windowScope.isActive = true;
					};
	
					// shows the window
					windowScope._show = function() {
						windowScope._load(function() {
							// get the window below this one
							var windowIndex = self.windows.indexOf(windowScope);
							var marginTopReduce = 0;
			
							if (windowIndex > 0) {
								var parentWnd = self.windows[windowIndex-1].dom;
								marginTopReduce = -1 * (parentWnd.outerHeight() - Math.max(80, Math.min($(document).scrollTop(), parentWnd.outerHeight()-80)));
							}
	
							// append us to the dom
							$('#edit-windows').append(windowScope.dom);
			
							windowScope.activate();
			
							// set the width
							windowScope.dom.css('width', (720 + 30*self.windows.length - 30/self.windows.length) + 'px');
			
							// wait a short time before animating, because we want the dom to settle
							$timeout(function() {
								windowScope.dom.addClass('actives').animate({
									'margin-top': marginTopReduce + 'px',
									'opacity': 1
								}, (windowScope.nested ? 400 : 200));
							}, 50);
							
							$rootScope.reactivateModals();							
						});
					};

					//
					// SHOW US
					//
					if (nested || self.windows.length === 1) {
						// show ourself when nested
						windowScope._show();
					} else {
						// close all other windows if not nested and wait for the last one to close
						for(var i=0; i<self.windows.length; i++) {
							var wnd = self.windows[i];
							if (wnd !== windowScope) {
								wnd.close(function() {
									if (self.windows.length === 1) {
										windowScope._show();
									}
								}, true);
							}
						}
					}
				}, 10);
			}
		};
	}
	return window.windowHandler;
}]);

/**
 * Media Service
 * Provides functionality to show and pick items from the Media Gallery
 */
angular.module('app').service('media', ['modal', function(modal) {
	this.select = function(callback, initial) {
		modal.show('MediaGalleryModal', function(res) {
			if (res && typeof callback === 'function') {
				callback(res);
			}
		}, 
		// injects
		{
			"initial" : initial ? initial : false,
			"selectable" : typeof(callback) === 'function'
		});
	};
}]);

/**
 * Datasource Service
 * Provides Functionality to get Data for a given Datasource via Ajax.
 * Also Caches already fetched Data.
 */
angular.module('app').service('datasource', ['message', 'core', function(message, core) {

	var self = this;
	this._cache = {};
	this._resolveCache = {};

	this.get = function(datasource, callback, nocache) {

		if (angular.isString(datasource)) {
			// lookup in cache
			if (!nocache && this._cache[datasource]) {
				setTimeout(function() {
					callback(self._cache[datasource]);
				});
				return;
			}
			// live call
			else {
				$.ajax({
					type: 'POST',
					dataType: 'JSON',
					url: core.shopURL,
					data: {
						"kk_get_datasource": datasource,
						"kk_dropper_admin_pass": core.adminSecret
					},
					success: function(res) {

						if (!res.results) { 
							message.error('Error fetching Datasource "' + datasource + '"', res, false);
							callback(null);
							return;
						}

						if (res.cacheable) self._cache[datasource] = res.results;
						return callback(res.results);

					},
					error: function(err) {
						message.error('Error fetching Datasource "' + datasource + '"', err, false);
						callback(null);
					}
				});
				return;
			}
		}
		else {
			// we should never get to this point, hopefully
			message.error('Given Datasource was not in String Format:', datasource);
			setTimeout(function() {
				callback(null);
			});
		}
	};
	
	this.search = function(datasource, query, page, callback) {

		if (!angular.isString(datasource) || !angular.isString(query)) {

			// we should never get to this point, hopefully
			message.error('Given Datasource or query was not in String Format:', datasource);

			setTimeout(function() {
				callback(null);
			});

			return null;
		}
		
		return $.ajax({
			type: 'POST',
			url: core.shopURL,
			data: {
				"kk_get_datasource_search": datasource,
				"kk_get_datasource_search_query": query,
				"kk_get_datasource_search_page": page || 0,
				"kk_dropper_admin_pass": core.adminSecret
			},
			success: function(res) {
				if (res && res.results) {
					return callback(res);
				}
			},
			error: function(err) {
				// ignore errors
			}
		});
	};

	this.resolve = function(datasource, mapping, keys, callback) {

		// check for valid parameters
		if (!angular.isString(datasource) || !angular.isString(mapping)) {
			message.error('Could not resolve Datasource');
			setTimeout(function() { callback(null);	});
			return;
		}

		// default to array
		keys = keys || [];

		// resolve cache key
		var dsCacheKey = datasource + ':' + mapping;

		// keys that are being requested, because not in cache already
		var askKeys = [];
		
		// options that have been resolved, merged with cached items
		var finalOptions = [];
		
		// lookup cached items
		for (var i=0; i<keys.length; i++) {
			var key = keys[i];
			if (!self._resolveCache[dsCacheKey] || !self._resolveCache[dsCacheKey][key]) { 
				// not resolved yet, push it to the question
				askKeys.push(key);
			} else {
				// already resolved, push it to the final results
				finalOptions.push(self._resolveCache[dsCacheKey][key]);
			}
		}

		// if no keys to ask are left we either got all results from cache or there were no keys
		if (!askKeys.length) {

			setTimeout(function() {	callback(finalOptions);	});
			return;

		} 
		// make the ajax request with the remaining keys
		else {
			$.ajax({
				type: 'POST',
				url: core.shopURL,
				data: {
					"kk_get_datasource_query": datasource,
					"kk_get_datasource_query_mapping": mapping,
					"kk_get_datasource_query_keys": askKeys || [],
					"kk_dropper_admin_pass": core.adminSecret
				},
				success: function(res) {
					
					if (res && res.results) {

						// cache results and add them to the final options
						for (var i=0; i<res.results.length; i++) {
							var result = res.results[i];
							if (!self._resolveCache[dsCacheKey]) self._resolveCache[dsCacheKey] = {};
							if (typeof(result[mapping]) !== 'undefined') self._resolveCache[dsCacheKey][result[mapping]] = result;
							finalOptions.push(result);
						}
						
						// we are done!
						return callback(finalOptions);
					}
				},
				error: function(err) {
					
					setTimeout(function() {
						callback(null);
					});
				}
			});
		}
	
	};
	

}]);

/**
 * Utils Service
 * Provides functionality that might be usefull elsewhere
 */
angular.module('app').service('utils', ['environment', function(environment) {

	var self = this;

	/*************************
	 * UTF8 Conversion Util
	 *************************/
	this.Utf8 = {
	 
		// public method for url encoding
		encode : function (string) {
			string = string.replace(/\r\n/g,"\n");
			var utftext = "";
	 
			for (var n = 0; n < string.length; n++) {
	 
				var c = string.charCodeAt(n);
	 
				if (c < 128) {
					utftext += String.fromCharCode(c);
				}
				else if ((c > 127) && (c < 2048)) {
					utftext += String.fromCharCode((c >> 6) | 192);
					utftext += String.fromCharCode((c & 63) | 128);
				}
				else {
					utftext += String.fromCharCode((c >> 12) | 224);
					utftext += String.fromCharCode(((c >> 6) & 63) | 128);
					utftext += String.fromCharCode((c & 63) | 128);
				}
			}

			return utftext;
		},

		// public method for url decoding
		decode : function (utftext) {
			var string = "";
			var i = 0;
			var c, c2, c3;
			c = c2 = c3 = 0;
	 
			while ( i < utftext.length ) {
	 
				c = utftext.charCodeAt(i);
	 
				if (c < 128) {
					string += String.fromCharCode(c);
					i++;
				}
				else if ((c > 191) && (c < 224)) {
					c2 = utftext.charCodeAt(i+1);
					string += String.fromCharCode(((c & 31) << 6) | (c2 & 63));
					i += 2;
				}
				else {
					c2 = utftext.charCodeAt(i+1);
					c3 = utftext.charCodeAt(i+2);
					string += String.fromCharCode(((c & 15) << 12) | ((c2 & 63) << 6) | (c3 & 63));
					i += 3;
				}

			}

			return string;
		}
	};

	/*************************************
	 * Language Util
	 ************************************/
	this.Lang = {
		toMultilang: function(text) {
			if (angular.isString(text) || (angular.isObject(text) && typeof(text['_multilanguage']) === 'undefined')) {
				var mlText = {
					_multilanguage: true,
				};
				mlText[environment.languageDefault.cISO] = text;
				return mlText;
			}
			return text;
		}
	};

	/*************************************
	 * Drops Util
	 ************************************/
	this.Drops = {
		setDefaultConfig: function(item, defaultsObj, newName) {
			var defaults = defaultsObj || false;
			var name = newName || 'Neuer Eintrag';
			
			if (item) {
				var itemBefore = $.extend(true, {}, item);
				var preset = { 
					_internal: { 
						visibility: 'active', 
						name: name,
						guid: self.generateGuid() 
					}
				};
				
				if (defaults) {
					$.extend(true, item, preset, defaults, itemBefore);
				} else {
					$.extend(true, item, preset, itemBefore);
				}
			}
			return item;
		},
		
		addItem: function(list, itemDefaults, itemSettings, parent, newName) {
			var newItem = {
				_internal: {}
			};
	
			this.setDefaultConfig(newItem, itemDefaults || false, newName);
			list.push(newItem);

			return newItem;
		},

		deepReplaceGuids: function(obj) {
			// replace _internal.guid
			if (angular.isObject(obj)) {
				if (obj._internal && obj._internal.guid) {
					var newGuid = self.generateGuid();
					obj._internal.guid = newGuid;
				}
			}
			// recursively follow paths
			if (angular.isArray(obj) || angular.isObject(obj)) {
				for(var key in obj) {
					this.deepReplaceGuids(obj[key]);
				}
			}
		},

		copyItem: function(list, item, itemSettings, parent) {
			// duplicate
			var newItem = $.extend(true, {}, item);
			// don't copy the hash!
			if (newItem['$$hashKey']) {
				delete newItem['$$hashKey'];
			}

			newItem._internal.name += " (Kopie)";
			this.deepReplaceGuids(newItem);
			
			list.push(newItem);
			return newItem;
		}
	};

	/**
	 * Generates a GUID
	 */
	this.generateGuid = function() {
		return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
			var r = Math.random()*16|0, v = c == 'x' ? r : (r&0x3|0x8);
			return v.toString(16);
		});
	};
	
	/**
	 * Generates a smarty expression from the given conditional settings of the given drop
	 */
	this.conditionToSmartyExpression = function(conditionalSettings) {

		// empty settings? return empty expression
		if (!angular.isArray(conditionalSettings) || !conditionalSettings.length) return "";

		var exp = "";
		var error = false;
		var conditional = window.conditional;
	
		for(var i=0; i<conditionalSettings.length; i++) {
			var currentCond = conditionalSettings[i];
			var partExp = "";
	
			partExp = conditional[currentCond.conditional].evaluateToSmarty(currentCond.setting.choices);
	
			// in case the evaluateToSmarty method returns false, we
			// know there was an error... don't let it pass by (remove the Conditonal from the list as well)
			if (partExp === false) {
				conditionalSettings.splice(i, 1);
				i--;
				error = true;
			}
			else {
				if (currentCond.equality === false) {
					// inverted equality (is not)
					exp += "!(" + partExp + ")";
				} else {
					// equality (is)
					exp += "(" + partExp + ")";
				}
			}
	
			if (i < conditionalSettings.length-1 && !error) {
				exp += " " + currentCond.setting.operand + " ";
			}
		}
		if (error) {
			// there was an error within some condition. 
			// We removed it and run this stuff again
			return this.conditionToSmartyExpression(conditionalSettings);
		} else {
			return exp;
		}
	};

	/**
	 * Helper to regenerate conditions of all drops (and list items) from thier conditionals
	 */
	this.refreshConditionsDeep = function(obj, key) {
		if ( 
			angular.isObject(obj) && angular.isDefined(obj.conditionalSettings) && angular.isDefined(obj.condition) && 
			(key === '_internal' || obj['_conditional'] === true)
		) {
			obj.condition = this.conditionToSmartyExpression(obj.conditionalSettings);
		}

		// traverse children
		if (angular.isObject(obj)) {
			for (var key in obj) {
				this.refreshConditionsDeep(obj[key], key);
			}
		}
	};
}]);


/**
 * Upgrade Service that is able to upgrade Drop Data.
 * There are two different scenarios: Dropper Upgrades and
 * Drop upgrades form the upgrade.js file
 */
angular.module('app').service('upgrade', ['message', 'core', 'environment', function(message, core, environment) {

	if (window.kkUpgrader) return window.kkUpgrader;

	var dropUpgrader = {
		_dropperUpgrades: [],
		_dropUpgrades: [],
		
		/**
		 * Unregisters all upgrades
		 */
		cleanup: function() {
			this._dropUpgrades = [];
			this._dropperUpgrades = [];
		},

		/**
		 * Adds a new upgrade for the control center.
		 * @from - the version from which to upgrade (e.g. "100.1")
		 * @to - the version to which to upgrade (e.g. "100.2")
		 * @upgradeFn - the upgrade function, params: settings, templates
		 */
		addDropperUpgrade: function(from, to, upgradeFn) {
			console.log('adding dropper upgrade', arguments);
			this._dropperUpgrades.push({
				from: from,
				to: to,
				fn: upgradeFn
			});
		},

		/**
		 * Recursively runs all upgrades on the given data
		 * @settings - the Dropper settings object
		 * @returns - the upgraded settings object
		 */
		runDropperUpgrades: function(settings, cb /* changedSettings, count */) {
			if (!angular.isFunction(cb)) return console.log('Attempting to run Dropper Upgrades, but no callback given!');

			console.log('running dropper upgrades', this);
			return this._runDropperUpgrades(settings, 0, cb);
		},

		/**
		 * Recursively runs all upgrades on the given data
		 * @settings - the Dropper settings object
		 * @count - the Count of upgrades that where run
		 * @returns - the upgraded settings object
		 */
		_runDropperUpgrades: function(settings, count, cb) {
			var clonedSettings = $.extend(true, {}, settings);

			if (!count) {
				count = 0;
			}

			if (!clonedSettings.dropperVersion) {
				clonedSettings.dropperVersion = "100.0";
			}

			for(var i=0; i<this._dropperUpgrades.length; i++) {
				var upgrade = this._dropperUpgrades[i];
				
				if (clonedSettings.dropperVersion === upgrade.from) {
					// count one up
					count ++;
					console.log("Upgrading Dropper", upgrade.from, upgrade.to);
					var up = upgrade.fn(clonedSettings, core.templates);

					// recursively upgrade further
					if (!up || typeof(up) === 'string') {
						// error at last upgrade
						if (!up) up = 'Das Migrationsresultat war leer.';
						return cb('Das Dropper Upgrade von ' + upgrade.from + ' auf ' + upgrade.to + ' ist fehlgeschlagen: ' + up, count);
					}

					up.dropperVersion = upgrade.to;
					return this._runDropperUpgrades(up, count, cb);
				}
			}

			return cb(settings, count);
		},

		/**
		 * Adds a new upgrade for an drop.
		 * @dropId - the ID of the drop
		 * @from - the version from which to upgrade (e.g. "1.0.0")
		 * @to - the version to which to upgrade (e.g. "1.0.1")
		 * @upgradeFn - the upgrade function, params: drop, template
		 */
		addDropUpgrade: function(dropId, from, to, upgradeFn) {
			console.log('adding drop upgrade', arguments);
			this._dropUpgrades.push({
				dropId: dropId,
				from: from,
				to: to,
				fn: upgradeFn
			});
		},

		/**
		 * Recursively runs all upgrades on the given data
		 * @settings - the Dropper settings object
		 * @returns - the upgraded settings object
		 */
		runDropUpgrades: function(settings, cb /* upgradedSettings, count */) {
			console.log('running drop upgrades', this);
			var upgradeState = [];

			for(var i=0; i<settings.drops.length; i++) {
				var drop = settings.drops[i], template;
				if (core.templates[drop._internal.dropId]) {
					template = core.templates[drop._internal.dropId];
				} else {
					continue;
				}

				// run upgrade on this drop
				this._runDropUpgrades(drop, template, 0, $.proxy(function(upgradedDrop, count) {

					// nothing upgraded?
					if (count === 0) return;

					// upgrade success?
					if (upgradedDrop && !angular.isString(upgradedDrop)) {
						settings.drops[this.index] = upgradedDrop;
						upgradeState.push({
							drop: drop,
							error: false,
							count: count,
							template: template
						});
					} 
					// error
					else {
						upgradeState.push({
							drop: drop,
							error: upgradedDrop,
							count: count,
							template: template
						});
					}

				}, { index: i }));
			}

			return cb(settings, upgradeState);
		},


		/**
		 * Recursively runs all upgrades on the given data
		 * @drop - the drop data to upgrade
		 * @count - a number of upgrades already run on the data
		 * @returns - the upgraded settings object
		 */
		_runDropUpgrades: function(drop, template, count, cb) {
			var clonedDrop = $.extend(true, {}, drop);

			if (!count) {
				count = 0;
			}

			if (!clonedDrop._internal.version) {
				clonedDrop._internal.version = "1.0.0";
			}

			for(var i=0; i<this._dropUpgrades.length; i++) {
				var upgrade = this._dropUpgrades[i];
				
				if (clonedDrop._internal.dropId === upgrade.dropId && clonedDrop._internal.version === upgrade.from) {
					// count one up
					count ++;
					var up = upgrade.fn(clonedDrop, template);
					console.log("Drop Upgraded: ", upgrade.from, upgrade.to, up);

					// recursively upgrade further
					if (!up || typeof(up) === 'string') {
						// error at last upgrade
						if (!up) up = "Das Migrationsresultat war leer.";
						return cb('Das Upgrade des Drops "' + drop._internal.name + '" von ' + upgrade.from + ' auf ' + upgrade.to + ' ist fehlgeschlagen: ' + up, count);
					}

					up._internal.version = upgrade.to;
					return this._runDropUpgrades(up, template, count, cb);
				}
			}

			return cb(drop, count);
		},

		/**
		 * Check if there is a drop option defined for the given template ID
		 */
		dropHasOption: function(templateId, optionName) {
			var templates = core.templates;
			if (typeof templates === 'object' && typeof templates[templateId] === 'object') {
				var template = templates[templateId];
				if (template.config instanceof Array) {
					for (var i=0; i<template.config.length; i++) {
						if (template.config[i].property === optionName) return true;
					}
				}
				if (template.settings instanceof Array) {
					for (var i=0; i<template.settings.length; i++) {
						if (template.settings[i].property === optionName) return true;
					}
				}
			}
			return false;
		},

		/**
		 * Check if there is a dropItem option defined for the given template
		 */
		dropItemHasOption: function(templateId, optionName) {
			var templates = core.templates;
			if (typeof templates === 'object' && typeof templates[templateId] === 'object') {
				var template = templates[templateId];
				if (template.itemSettings instanceof Array) {
					for (var i=0; i<template.itemSettings.length; i++) {
						if (template.itemSettings[i].property == optionName) {
							return true;
						}
					}
				}
			}
			return false;
		},

		/**
		 * Iterates over a string type object and calls the callback function which in turn can edit
		 * split out text parts. This is very usefull for multilanguage fields.
		 */
		editText: function(obj, fn /* text, iso */) {
			if (angular.isString(obj)) {
				// simple string
				return fn(obj, environment.languageDefault.cISO);
			}
			if (angular.isObject(obj)) {
				// mutlilang object?
				if (obj['_multilanguage']) {
					var cloned = $.extend(true, {}, obj);
					var langCounter = 0;
					for(var lang in obj) {
						if (lang !== '_multilanguage') {
							cloned[lang] = fn(cloned[lang], lang);
							langCounter ++;
						}
					}
					// maybe the object looked like {_multilanguage: true} but no content set
					// handle it
					if (langCounter === 0) {
						cloned[environment.languageDefault.cISO] = fn("", environment.languageDefault.cISO);
					}
					return cloned;
				}
				// no multilang object
				else {
					return obj;
				}
			}
			if (typeof(obj) === "undefined" || obj === null) {
				// not set yet... but maybe you want to?
				return fn(null, environment.languageDefault.cISO);
			}
			// default
			return obj;
		},
		urlToLink: function(url, title) {
			return this.editText(url, function(url, iso) {
				return {
					_url: true,
					value: url,
					type: "Manual",
					title: (typeof(title) === "string" ? title : (angular.isObject(title) && angular.isString(title[iso]) ? title[iso] : ""))
				};
			});		
		}
	};

	/**
	 * This will return the upgrader
	 */
	window.kkUpgrader = dropUpgrader;
	return window.kkUpgrader;
}]);
